import re

class QuestException(Exception): pass
class StreakBroken(QuestException): pass
class QuestComplete(QuestException): pass
class QuestProgress(QuestException): pass

class Quest(object):
	goal = 5
	title = "quest"
	desc = "a quest!"
	def __init__(self, progress=0, completed=False):
		self.progress = progress
		self.completed = completed
	def word_formed(self, word, score, mult, drag): pass
	def power_up_collected(self, effect): pass
	def get_progress(self): return self.progress
	def get_goal(self): return self.goal
	def is_completed(self): return self.completed

class SevenPlusLettersThreeInARowQuest(Quest):
	goal = 3
	title = "Streak breaker"
	desc = "Make three words in a row with 7+ letters"

	def word_formed(self, word, score, mult, drag):
		if len(word) >= 7:
			self.progress += 1
			if self.progress == self.goal:
				self.completed = True
				raise QuestComplete
			else:
				raise QuestProgress
		elif self.progress > 0:
			self.progress = 0
			raise StreakBroken

class SixLettersThreeInARowQuest(Quest):
	goal = 3
	title = "Six times three"
	desc = "Make three 6-letter words in a row"

	def word_formed(self, word, score, mult, drag):
		if len(word) == 6:
			self.progress += 1
			if self.progress == self.goal:
				self.completed = True
				raise QuestComplete
			else:
				raise QuestProgress
		elif self.progress > 0:
			self.progress = 0
			raise StreakBroken

class FiveLettersThreeInARowQuest(Quest):
	goal = 3
	title = "Five times three"
	desc = "Make three 5-letter words in a row"

	def word_formed(self, word, score, mult, drag):
		if len(word) == 5:
			self.progress += 1
			if self.progress == self.goal:
				self.completed = True
				raise QuestComplete
			else:
				raise QuestProgress
		elif self.progress > 0:
			self.progress = 0
			raise StreakBroken
			self

class SixIn23BoxQuest(Quest):
	goal = 1
	title = "Six Box"
	desc = "Make a 6-letter word in a 2x3 or 3x2 box"

	def word_formed(self, word, score, mult, drag):
		hw_tup = (drag.get_width(), drag.get_height())
		if len(word) == 6 and hw_tup in ((2,3),(3,2)):
			self.progress += 1
			self.completed = True
			raise QuestComplete

class NineIn33BoxQuest(Quest):
	goal = 1
	title = "The matrix"
	desc = "Make a 9-letter word in a 3x3 box"

	def word_formed(self, word, score, mult, drag):
		hw_tup = (drag.get_width(), drag.get_height())
		if len(word) == 9 and hw_tup == (3,3):
			self.progress += 1
			self.completed = True
			raise QuestComplete

class FourLettersFourInARowQuest(Quest):
	goal = 4
	title = "Four by four" 
	desc = "Form four 4-letter words in a row"

	def word_formed(self, word, score, mult, drag):
		if len(word) == 4:
			self.progress += 1
			if self.progress == self.goal:
				self.completed = True
				raise QuestComplete
			else:
				raise QuestProgress
		elif self.progress > 0:
			self.progress = 0
			raise StreakBroken
			self

class FiveLetterStraightQuest(Quest):
	goal = 1
	title = "Straight five"
	desc = "Make a 5-letter word with tiles in a straight line"
	
	def word_formed(self, word, score, mult, drag):
		h = drag.get_height()
		w = drag.get_width()
		if len(word) == 5 and ((h == len(word) and w == 1) or \
				(w == len(word) and h == 1)):
			self.progress += 1
			self.completed = True
			raise QuestComplete

class SixLetterStraightQuest(Quest):
	goal = 1
	title = "Six shooter"
	desc = "Make a 6-letter word with tiles in a straight line"
	
	def word_formed(self, word, score, mult, drag):
		h = drag.get_height()
		w = drag.get_width()
		if len(word) == 6 and ((h == len(word) and w == 1) or \
				(w == len(word) and h == 1)):
			self.progress += 1
			self.completed = True
			raise QuestComplete

class OneSixLetterWordQuest(Quest):
	goal = 1
	title = "Single six"
	desc = "Make a 6-letter word"

	def word_formed(self, word, score, mult, drag):
		if len(word) == 6:
			self.progress += 1
			self.completed = True
			raise QuestComplete

class OneSevenLetterWordQuest(Quest):
	goal = 1
	title = "Seven up"
	desc = "Make a 7-letter word"

	def word_formed(self, word, score, mult, drag):
		if len(word) == 7:
			self.progress += 1
			self.completed = True
			raise QuestComplete

class OneEightLetterWordQuest(Quest):
	goal = 1
	title = "Octogram"
	desc = "Make an 8-letter word"

	def word_formed(self, word, score, mult, drag):
		if len(word) == 8:
			self.progress += 1
			self.completed = True
			raise QuestComplete

class OneNineLetterWordQuest(Quest):
	goal = 1
	title = "Cloud nine"
	desc = "Make a 9-letter word"

	def word_formed(self, word, score, mult, drag):
		if len(word) == 9:
			self.progress += 1
			self.completed = True
			raise QuestComplete

class OneTenLetterWordQuest(Quest):
	goal = 1
	title = "Perfect"
	desc = "Make a 10-letter word"

	def word_formed(self, word, score, mult, drag):
		if len(word) == 10:
			self.progress += 1
			self.completed = True
			raise QuestComplete

class JQXZWordQuest(Quest):
	goal = 1
	title = "Quick Jazz Sax Zinger"
	desc = "Make a word that contains J, Q, X, or Z"

	def word_formed(self, word, score, mult, drag):
		if re.search(r'[jqxz]', word):
			self.progress += 1
			self.completed = True
			raise QuestComplete

class ThreePowerupsInARowQuest(Quest):
	goal = 3
	title = "Troika"
	desc = "Collect three power-ups without forming a word"

	def word_formed(self, word, score, mult, drag):
		if self.progress > 0:
			self.progress = 0
			raise StreakBroken

	def power_up_collected(self, effect):
		self.progress += 1
		if self.progress == self.goal:
			self.completed = True
			raise QuestComplete
		else:
			raise QuestProgress

class OneOfEachPowerupQuest(Quest):
	goal = 5
	title = "The Compleat Miner"
	desc = "Collect one of each power-up"

	def __init__(self, progress=0, completed=False):
		self.powerup_set = set()
		super(OneOfEachPowerupQuest, self).__init__(progress,
				completed)

	def power_up_collected(self, effect):
		if effect not in self.powerup_set:
			self.powerup_set.add(effect)
			self.progress += 1
			if len(self.powerup_set) >= 5:
				self.completed = True
				raise QuestComplete
			else:
				raise QuestProgress

class SameWordThreeTimesInARowQuest(Quest):
	goal = 3
	title = "Deja Vu"
	desc = "Form the same word of 4+ letters 3 times in a row"

	def __init__(self, progress=0, completed=False):
		self.word = None
		super(SameWordThreeTimesInARowQuest, self).__init__(progress, completed)

	def word_formed(self, word, score, mult, drag):
		if word == self.word:
			self.progress += 1
			if self.progress == self.goal:
				self.completed = True
				raise QuestComplete
			else:
				raise QuestProgress
		else:
			if len(word) > 3 and self.word is None:
				self.word = word
				self.progress = 1
				raise QuestProgress
			else:
				self.word = None
				self.progress = 0
				raise StreakBroken

class ThreeInARowSameFirstLetterQuest(Quest):
	goal = 3
	title = "Alliteration"
	desc = "Form 3 words in a row with same first letter"

	def __init__(self, progress=0, completed=False):
		self.init_letter = None
		super(ThreeInARowSameFirstLetterQuest, self).__init__(progress,
				completed)
	def word_formed(self, word, score, mult, drag):
		if word[0] == self.init_letter:
			self.progress += 1
			if self.progress == self.goal:
				self.completed = True
				raise QuestComplete
			else:
				raise QuestProgress
		else:
			if self.init_letter is None:
				self.init_letter = word[0]
				self.progress = 1
				raise QuestProgress
			else:
				self.init_letter = None
				self.progress = 0
				raise StreakBroken

class FivePlusOnlyDiagonalDragsQuest(Quest):
	goal = 1
	title = "Kitty corner"
	desc = "Form a word with 5+ letters dragging diagonally only"

	def word_formed(self, word, score, mult, drag):
		if len(word) < 5: return
		positions = drag.get_positions()
		last_pos = positions[0]
		for pos in positions[1:]:
			if abs(pos[0]-last_pos[0]) != 1 or abs(pos[1]-last_pos[1]) != 1:
				return
			last_pos = pos
		self.progress = 1
		self.completed = True
		raise QuestComplete

class FivePlusOnlyUpwardDragsQuest(Quest):
	goal = 1
	title = "Nowhere to go"
	desc = "Form a word with 5+ letters dragging upward (N, NE, NW) only"

	def word_formed(self, word, score, mult, drag):
		if len(word) < 5: return
		positions = drag.get_positions()
		last_pos = positions[0]
		for pos in positions[1:]:
			if pos[1] >= last_pos[1]: return
			last_pos = pos
		self.progress = 1
		self.completed = True
		raise QuestComplete

class SevenPlusLettersOneVowelQuest(Quest):
	goal = 1
	title = "Loneliest vowel"
	desc = "Word with 7+ letters but only one vowel"

	def word_formed(self, word, score, mult, drag):
		if len(word) < 7: return
		vowels = [c for c in word if c in 'aeiou']
		if len(vowels) == 1:
			self.progress = 1
			self.completed = True
			raise QuestComplete

class FourPlusLettersOneConsonantQuest(Quest):
	goal = 3
	title = "Vocal harmony"
	desc = "3 words with 4+ letters but only 1 consonant"

	def word_formed(self, word, score, mult,drag):
		if len(word) < 4: return
		consonants = [c for c in word if c not in 'aeiou']
		if len(consonants) == 1:
			self.progress += 1
			if self.progress == self.goal:
				self.completed = True
				raise QuestComplete
			else:
				raise QuestProgress

class VKWZXQJQuest(Quest):
	goal = 1
	title = "Rare gems"
	desc = "1 word with 2+ of V, K, W, Z, X, Q, J"

	def word_formed(self, word, score, mult, drag):
		if re.search(r'[vkwzxqj].*[vkwzxqj]', word):
			self.progress += 1
			self.completed = True
			raise QuestComplete

class AlphabetQuest(Quest):
	goal = 3
	title = "In its right place"
	desc = "3 words whose 4+ letters are in alphabetical order"

	def word_formed(self, word, score, mult, drag):
		if len(word) < 4: return
		if word == ''.join(sorted(list(word))):
			self.progress += 1
			if self.progress == self.goal:
				self.completed = True
				raise QuestComplete
			else:
				raise QuestProgress

class QuestManager(object):

	def __init__(self, score_manager):
		self.score_manager = score_manager
		self.user_id = score_manager.user_id
		self.quests = [[],[],[],[],[],[]]

	def initialize(self):
		quests = [
			[FourLettersFourInARowQuest, FiveLettersThreeInARowQuest,
					FiveLetterStraightQuest, OneSixLetterWordQuest],
			[SixLettersThreeInARowQuest, JQXZWordQuest, OneSevenLetterWordQuest,
					ThreePowerupsInARowQuest],
			[ThreeInARowSameFirstLetterQuest, SixLetterStraightQuest,
					OneEightLetterWordQuest, OneOfEachPowerupQuest],
			[SixIn23BoxQuest, SevenPlusLettersThreeInARowQuest,
					FivePlusOnlyDiagonalDragsQuest, OneNineLetterWordQuest],
			[FivePlusOnlyUpwardDragsQuest, FourPlusLettersOneConsonantQuest,
					OneTenLetterWordQuest, AlphabetQuest],
			[SameWordThreeTimesInARowQuest, VKWZXQJQuest, NineIn33BoxQuest,
					SevenPlusLettersOneVowelQuest],
		]
		db = self.score_manager.quests
		for level, level_list in enumerate(quests):
			for quest in level_list:
				db.insert(user_id=self.user_id, name=quest.__name__, level=level,
						progress=0, complete=False)

	def restore(self):
		db = self.score_manager.quests
		for quest in db(user_id=self.user_id):
			inst = globals()[quest.name](progress=quest.progress,
					completed=quest.complete)
			self.quests[quest.level].append(inst)

	def word_formed(self, word, score, mult, drag, callback):
		to_save = list()
		current_level = self.get_current_level()
		if current_level is None: return
		for quest in (q for q in self.quests[current_level] \
				if not q.is_completed()):
			try:
				quest.word_formed(word, score, mult, drag)
			except QuestException, e:
				to_save.append(quest)
				callback(quest, e)
		for quest in to_save:
			self.update_quest(quest)

	def power_up_collected(self, effect, callback):
		to_save = list()
		current_level = self.get_current_level()
		if current_level is None: return
		for quest in (q for q in self.quests[current_level] \
				if not q.is_completed()):
			try:
				quest.power_up_collected(effect)
			except QuestException, e:
				to_save.append(quest)
				callback(quest, e)
		for quest in to_save:
			self.update_quest(quest)

	def update_quest(self, quest):
		rec = self.score_manager.quests.select_for_update(
			name=quest.__class__.__name__)
		self.score_manager.quests.update(
			rec,
			progress=quest.progress,
			complete=quest.is_completed())

	def get_quests(self):
		return self.score_manager.quests(user_id=self.user_id)

	def get_current_level(self):
		for i, level in enumerate(self.quests):
			if not all([q.is_completed() for q in level]): return i

	def completion_tuple(self):
		return (len([1 for lev in self.quests for q in lev \
				if q.is_completed() is True]),
				len([q for lev in self.quests for q in lev]))

	def all_complete(self):
		return all([q.is_completed() for lev in self.quests for q in lev])

